from typing import Optional

from popupdict.config import *
from popupdict.query.result import QueryResult
from .abstract import AbstractQueryClient


# 使用有道词典网页版进行查词翻译
class YoudaoWebQueryClient(AbstractQueryClient):
    def __init__(self, config: YoudaoWebConfiguration):
        super().__init__(config)
        self.config = config

    def query(self, word: str) -> Optional[QueryResult]:
        raise Exception('Client not implemented!')



from .abstract import AbstractQueryClient
from .fake import FakeQueryClient
from .youdao_web import YoudaoWebQueryClient
from .youdao_zhiyun import YoudaoZhiyunQueryClient

__all__ = [
    'AbstractQueryClient',
    'FakeQueryClient',
    'YoudaoWebQueryClient',
    'YoudaoZhiyunQueryClient',
]



import requests
import random
import hashlib
import json
from typing import Optional

from popupdict.config import *
from popupdict.query.result import QueryResult
from .abstract import AbstractQueryClient


# 使用有道智云的自然语言翻译服务进行查词翻译
class YoudaoZhiyunQueryClient(AbstractQueryClient):
    API = 'https://openapi.youdao.com/api'
    DICT_LINK = 'https://dict.youdao.com/w/{}/{}/'

    def __init__(self, config: YoudaoZhiyunConfiguration):
        super().__init__(config)
        self.config = config

    def query(self, text: str) -> Optional[QueryResult]:
        try:
            res = requests.get(self.API, params=self._params(text), timeout=self.config.request_timeout)
            if not res.ok:
                print("Request failed for text={}: status_code={}, reason={}, content={}".format(
                    repr(text), res.status_code, res.reason, res.text))
                return None

            try:
                result = res.json()
            except json.decoder.JSONDecodeError as e:
                print("Invalid JSON response for text={}: ".format(repr(text), res.text))
                return

            if result['errorCode'] != '0':
                print("Query failed for word={}: errorCode={}".format(
                    repr(text), str(result['errorCode'])))
                return None

            language = result['l'].replace('2zh-CHS', '').lower()

            basic = result.get('basic')
            phrases = result.get('web')
            if phrases is not None:
                for phrase in phrases:
                    phrase['value'] = '; '.join(phrase['value'])
                    phrase['dict_link'] = __class__.dict_link(language, phrase['key'])

            # 小语种查询结果没有 'query'
            query_result = QueryResult(result.get('query', text), '; '.join(result['translation']),
                                       dict_link=__class__.dict_link(language, result.get('query', text)),
                                       phonetic=basic and (
                                           basic.get('us-phonetic') or
                                           basic.get('uk-phonetic') or
                                           basic.get('phonetic')),
                                       explanations=basic and basic.get('explains'),
                                       phrases=phrases
                                       )
            return query_result
        except requests.exceptions.RequestException as e:
            print("Request failed for text={}: {}".format(repr(text), repr(e)))

    @staticmethod
    def dict_link(lang, query):
        return __class__.DICT_LINK.format(lang, __class__.escape_url_path(query))

    # 构造请求参数
    def _params(self, text):
        # http://ai.youdao.com/docs/doc-trans-api.s#p02
        params = {
            'q': text,
            # 设置为 auto 会导致有时识别错误
            'from': 'EN',
            'to': 'zh-CHS',
            'appKey': self.config.app_id,
            'salt': random.randint(1, 10000),
        }
        s = self.config.app_id + text + str(params['salt']) + self.config.app_secret
        params['sign'] = hashlib.md5(s.encode()).hexdigest()
        return params



import time

from popupdict.gtk import *
from popupdict.query import QueryResult
from popupdict.util import Selection
from .widgets import Widgets


# 弹窗
class Popup(Gtk.Window):
    WINDOW_WIDTH = 400  # type: int

    def __init__(self, popup_timeout: float = None):
        super().__init__(type=Gtk.WindowType.POPUP)

        # 弹窗显示时间。单位：秒
        self.popup_timeout = popup_timeout or 3  # type: float

        self.set_border_width(10)
        self.set_default_size(__class__.WINDOW_WIDTH, -1)
        self.set_size_request(__class__.WINDOW_WIDTH, -1)

        # Initialize container, widgets
        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, halign=Gtk.Align.START)
        self.widgets = Widgets(self.box)
        self.add(self.box)

        # Time to hide to window. Timestamp with float point got by time.time()
        self.time_to_hide = 0  # type: float

        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
        clipboard.connect("owner-change", __class__.selection_changed)

    # 渲染翻译结果
    def redraw(self, selection: Selection, query_result: QueryResult):
        # 若选中文本已经变化，不再显示旧的查询结果
        if not query_result or selection.text != Selection.current.text:
            return
        self.widgets.draw(query_result)
        minimum_height, natural_height = self.box.get_preferred_height()
        # Avoid warning: Allocating size to window without calling gtk_widget_get_preferred_width/height()
        # self.get_preferred_width()
        # self.get_preferred_height()
        self.resize(self.WINDOW_WIDTH, minimum_height)
        self.move_window(selection)
        self.show()
        self.time_to_hide = time.time() + self.popup_timeout

    # 根据选中文本位置、弹窗大小、屏幕大小确定弹窗位置，保证弹窗显示完整
    def move_window(self, selection: Selection):
        target_x, target_y = selection.position
        monitor_size = selection.monitor.get_geometry()
        width, height = self.get_size()

        def try_move(x, y):
            if x >= 0 and y >= 0 and x + width <= monitor_size.width and y + height <= monitor_size.height:
                self.move(x, y)
                return True
            return False

        # 下
        if try_move(target_x, target_y + 30):
            return
        # 上
        if try_move(target_x, target_y - height - 30):
            return
        # 左
        if try_move(target_x - width - 40, target_y - height / 2):
            return
        # 右
        if try_move(target_x + 40, target_y - height / 2):
            return
        self.move(10, 10)

    # 选中文本变化事件处理
    @staticmethod
    def selection_changed(clipboard: Gtk.Clipboard, event: Gdk.EventOwnerChange):
        text = clipboard.wait_for_text()
        Selection.update(text, event)



import os.path
from configparser import ConfigParser
from typing import List
import pkg_resources

from popupdict.gtk import *
from .clients import *


class Configuration:
    # 默认配置，始终加载
    DEFAULT_CONFIG = pkg_resources.resource_string(__package__, 'default.ini').decode()
    # 候选配置文件，按顺序检查，只加载第一个存在的文件
    CANDIDATE_CONFIG_FILES = [
        os.path.join(GLib.get_user_config_dir(), 'popup-dict/config.ini'),
        '/etc/popup-dict/config.ini',
    ]  # type: List[str]

    def __init__(self, config_file: str = None):
        parser = ConfigParser(default_section='client')
        parser.read_string(__class__.DEFAULT_CONFIG)

        if config_file:
            if not os.path.exists(config_file):
                raise ConfigError("Config file {} does not exist!".format(repr(config_file)))
            else:
                parser.read(config_file)
        else:
            # 若未指定配置文件，从候选配置文件中加载第一个存在的文件
            for path in __class__.CANDIDATE_CONFIG_FILES:
                if os.path.exists(path):
                    parser.read(path)
                    break

        if not parser.has_section('global'):
            raise ConfigError('Missing configuration section: global')

        # 全局配置
        try:
            global_section = parser['global']
            self.query_client = global_section['query_client']

            # 弹窗显示时间
            self.popup_timeout = global_section.getfloat('popup_timeout')

            # Gtk Global Dark Theme
            # 不设置或设为空则使用系统默认设置
            prefer_dark_theme = global_section.get('prefer_dark_theme')
            if prefer_dark_theme is not None and prefer_dark_theme != '':
                self.prefer_dark_theme = global_section.getboolean('prefer_dark_theme')
                Gtk.Settings.get_default().set_property("gtk-application-prefer-dark-theme", self.prefer_dark_theme)
            else:
                self.prefer_dark_theme = None

        except KeyError as e:
            raise ConfigError('Missing global configuration: ' + e.args[0])

        # 各查询客户端配置
        try:
            self.fake = FakeConfiguration()
            self.youdao_web = YoudaoWebConfiguration(parser['youdao-web'])
            self.youdao_zhiyun = YoudaoZhiyunConfiguration(parser['youdao-zhiyun'])
        except KeyError as e:
            raise ConfigError('Missing configuration section: ' + e.args[0])

    def __repr__(self):
        return 'Configuration(\n  query_method: {},\n  youdao_web: {},\n  youdao_zhiyun: {}\n)'.format(
            self.query_client, self.youdao_web, self.youdao_zhiyun)



import random

from .abstract import AbstractQueryClient
from popupdict.query.result import QueryResult


# 随机返回查询结果。开发/测试时使用，避免频繁调用接口
class FakeQueryClient(AbstractQueryClient):
    QUERY_RESULTS = [
        QueryResult('phonetic', '语音',
                    phonetic="fə'nɛtɪk",
                    explanations=['adj. 语音的，语音学的；音形一致的；发音有细微区别的'],
                    phrases=[{'value': '语音的; 语音学的; 语言的', 'key': 'phonetic'},
                             {'value': '语音特征; 语音要素; 语音面貌', 'key': 'phonetic feature'},
                             {'value': '注音; 直音注音; 标音', 'key': 'phonetic notation'}]
                    ),
        QueryResult('resume', '个人简历',
                    phonetic="rezju:'mei, ri'zju:m; -'zu:m",
                    explanations=['n. 摘要；[管理] 履历，简历', 'vi. 重新开始，继续', 'vt. 重新开始，继续；恢复，重新占用'],
                    phrases=[{'value': '重新开始; 简历; 恢复', 'key': 'resume'}, {'value': '简历', 'key': 'Resume'},
                             {'value': '视频简历; 视频简历; 频简历', 'key': 'Video resume'}]
                    ),
        QueryResult('fraught', '令人担忧的',
                    phonetic='frɔt',
                    explanations=['adj. 担心的，忧虑的；充满…的'],
                    phrases=[{'value': '误人子弟; 充满的; 忧虑的', 'key': 'Fraught'},
                             {'value': '伴随的', 'key': 'fraught a'},
                             {'value': '充满; 误人子弟; 带有', 'key': 'fraught with'}]
                    ),
    ]

    DICT_LINK = 'https://dict.youdao.com/w/eng/{}/'
    for query_result in QUERY_RESULTS:
        query_result.dict_link = DICT_LINK.format(AbstractQueryClient.escape_url_path(query_result.query))
        if query_result.phrases:
            for phrase in query_result.phrases:
                phrase['dict_link'] = DICT_LINK.format(AbstractQueryClient.escape_url_path(phrase['key']))

    def query(self, word: str) -> QueryResult:
        return __class__.QUERY_RESULTS[random.randint(0, len(__class__.QUERY_RESULTS) - 1)]



from typing import Optional
from configparser import SectionProxy


class ConfigError(Exception):
    pass


# Base class for configurations of various clients
class ClientConfiguration:
    def __init__(self, request_timeout=None):
        self.request_timeout = request_timeout or 3

    # Validate configuration, raise exception if invalid
    def validate(self):
        if not isinstance(self.request_timeout, int):
            raise ConfigError('Invalid timeout type: {}. Integer required!'.format(type(self.request_timeout)))
        elif self.request_timeout <= 0:
            raise ConfigError('Invalid timeout value: {}. Should be positive!'.format(self.request_timeout))

    def __repr__(self):
        result = self.__class__.__name__ + "("
        for k, v in self.__dict__.items():
            result += str(k) + '=' + str(v) + ', '
        result = result[:-2]
        result += ')'
        return result


class FakeConfiguration(ClientConfiguration):
    pass


class YoudaoWebConfiguration(ClientConfiguration):
    def __init__(self, section: SectionProxy):
        super().__init__(section.getint('request_timeout'))


class YoudaoZhiyunConfiguration(ClientConfiguration):
    def __init__(self, section: SectionProxy):
        super().__init__(section.getint('request_timeout'))
        self.app_id = section.get('app_id')  # type: Optional[str]
        self.app_secret = section.get('app_secret')  # type: Optional[str]

    def validate(self):
        super().validate()
        if self.app_id is None or len(self.app_id) == 0:
            raise ConfigError('app_id is empty!')
        elif self.app_secret is None or len(self.app_secret) == 0:
            raise ConfigError('app_secret is empty!')



import urllib.parse
from abc import ABC, abstractmethod
from typing import Optional

from popupdict.query.result import QueryResult
from popupdict.config import *


# Base class for query client
class AbstractQueryClient(ABC):
    def __init__(self, config: ClientConfiguration):
        config.validate()
        self.config = config

    @abstractmethod
    def query(self, word: str) -> Optional[QueryResult]:
        pass

    # Escape URL path segment
    @staticmethod
    def escape_url_path(s: str):
        return urllib.parse.quote(s, safe='')



from .configuration import Configuration
from .clients import *

__all__ = [
    'Configuration',
    'ConfigError',
    'ClientConfiguration',
    'FakeConfiguration',
    'YoudaoWebConfiguration',
    'YoudaoZhiyunConfiguration',
]



import time
import threading
from typing import Optional
from concurrent.futures import Future, ThreadPoolExecutor

from popupdict.gtk import *
from popupdict.ui import Popup
from popupdict.query import QueryAdapter, QueryResult
from popupdict.util import Selection


# 监视选中文本，如果变化，发起翻译请求，翻译完成后通知主线程渲染 UI
class QueryDaemon(threading.Thread):
    # 查询延时（选中文本多长时间不再变化时查询）。单位：秒
    QUERY_DELAY = 0.1

    def __init__(self, popup: Popup, query_adapter: QueryAdapter):
        super().__init__(daemon=True)
        self.popup = popup
        self.query_adapter = query_adapter
        self.executor = ThreadPoolExecutor(10)

    def run(self):
        query_task = None  # type: Optional[Future]
        last_selection = None  # type: Optional[Selection]
        while True:
            # 翻译完成，通知主线程渲染
            if query_task and query_task.done():
                query_result = query_task.result()  # type: QueryResult
                query_task = None
                if query_result:
                    # 只在主线程中渲染 UI
                    GLib.idle_add(self.popup.redraw, last_selection, query_result)

            # 选中内容不再变化时，发起翻译请求。避免选中文本过程中频繁请求
            selection = Selection.current
            if selection and not selection.queried and (
                    selection.time + __class__.QUERY_DELAY <= time.time() <=
                    selection.time + self.popup.popup_timeout):
                # 取消未完成的查询请求
                # TODO 如果请求已经在执行，Future.cancel 并不能真正取消
                if query_task:
                    query_task.cancel()

                last_selection = selection
                query_task = self.executor.submit(self.query_adapter.query, selection.text)
                selection.queried = True

            time.sleep(0.05)



from typing import Optional, List, Dict


# 查询结果
class QueryResult:
    def __init__(self, query: str, translation: str,
                 dict_link: Optional[str] = None,
                 phonetic: Optional[str] = None,
                 explanations: Optional[List[str]] = None,
                 phrases: Optional[List[Dict[str, str]]] = None):
        # 查询的单词
        self.query = query  # type: str
        # 在线词典链接
        self.dict_link = dict_link  # type: Optional[str]
        # 简略释义（一句话解释）
        self.translation = translation  # type: str
        # 音标
        self.phonetic = phonetic  # type: Optional[str]
        # 基本释义
        self.explanations = explanations  # type: Optional[List[str]]
        # 词组短语 [{"key": "词组/短语", value: ["翻译"], "dict_link": "在线词典链接（可为 None）"}]
        self.phrases = phrases  # type: Optional[List[Dict[str, str]]]

    def __repr__(self):
        return repr(self.__dict__)



from typing import Optional

from popupdict.config import *
from .client import *
from .result import QueryResult


# Query client adapter
class QueryAdapter:
    def __init__(self, config: Configuration):
        if config.query_client == 'fake':
            client = FakeQueryClient(config.fake)
        elif config.query_client == 'youdao-web':
            client = YoudaoWebQueryClient(config.youdao_web)
        elif config.query_client == 'youdao-zhiyun':
            client = YoudaoZhiyunQueryClient(config.youdao_zhiyun)
        else:
            raise Exception("Unknown query method: {}".format(repr(config.query_client)))

        self.client = client  # type: AbstractQueryClient

    def query(self, word: str) -> Optional[QueryResult]:
        try:
            result = self.client.query(word)
            return result
        except Exception as e:
            print("Query failed: {}".format(repr(e)))



import threading
import time

from popupdict.gtk import *
from popupdict.ui.popup import Popup


# 弹窗显示一段时间后，自动隐藏
# 为何使用单独的线程：
#    1. time.sleep 会阻塞整个线程，因此不宜在主线程中调用（拖慢更新选中文本的速度）
#    2. 若主线程正在阻塞（比如在等待 HTTP 请求完成），会无法隐藏窗口
class AutoHidePopupDaemon(threading.Thread):
    def __init__(self, popup: Popup):
        super().__init__(daemon=True)
        self.popup = popup
        # 鼠标设备
        self.pointer_device = Gdk.Display.get_default().get_default_seat().get_pointer()  # type: Gdk.Device

    # 检查鼠标是否在窗口
    def is_pointer_in_window(self) -> bool:
        _, pointer_x, pointer_y = self.pointer_device.get_position()
        x, y = self.popup.get_position()
        width, height = self.popup.get_size()
        return x < pointer_x < x + width and y < pointer_y < y + height

    def run(self):
        while True:
            if self.popup.is_visible() and time.time() >= self.popup.time_to_hide:
                # 若鼠标在窗口中，延迟隐藏窗口
                if not self.is_pointer_in_window():
                    # self.window.hide()
                    # 只在主线程中操作 UI
                    GLib.idle_add(self.popup.hide)
            time.sleep(0.3)



import re
from typing import Optional


# 过滤选中文本（删除两端的非字符内容、数字）
# 暂时只支持英文
class SelectionFilter:
    FILTER_PATTERN = re.compile('^[\W\d]*([a-zA-Z\s-]+)[\W\d]*$', re.ASCII)
    WHITESPACE_PATTERN = re.compile('\s+')

    @staticmethod
    def filter(text: str) -> Optional[str]:
        if not text or len(text) == 0 or len(text) > 80:
            return None

        # Remove soft hyphen
        # https://en.wikipedia.org/wiki/Soft_hyphen
        text = text.replace('\xad', '')

        # 主要处理合成词，暂不考虑 word break，（至少 PC 端）较少遇到
        # https://en.oxforddictionaries.com/punctuation/hyphen
        text = text.replace('-\n', '-')

        match = __class__.FILTER_PATTERN.fullmatch(text)
        if match:
            text = match.group(1)
        else:
            return None

        # 压缩空白字符（包括换行符），替换为一个空格
        text = __class__.WHITESPACE_PATTERN.sub(' ', text)
        if len(text) == 0:
            return None

        return text



import signal
import os.path
import sys
import psutil
import argparse
from typing import Optional

from .gtk import *
from .config import Configuration, ConfigError
from .query import QueryAdapter
from .ui import Popup
from .daemon import AutoHidePopupDaemon, QueryDaemon

PID_FILE = os.path.join(GLib.get_user_cache_dir(), 'popup-dict/popup-dict.pid')


def read_pid() -> Optional[int]:
    if os.path.exists(PID_FILE):
        with open(PID_FILE) as f:
            try:
                return int(f.read().strip())
            except ValueError:
                pass
    return None


def write_pid(pid: int):
    os.makedirs(os.path.dirname(PID_FILE), 0o755, True)
    with open(PID_FILE, 'w') as f:
        f.write(str(pid))


def delete_pid():
    if os.path.exists(PID_FILE):
        os.remove(PID_FILE)


def is_running():
    pid = read_pid()
    if pid and psutil.pid_exists(pid):
        process = psutil.Process(pid)
        if 'dict' in process.name():
            return True
    return False


def start(config_file: str = None, debug: bool = False):
    config = Configuration(config_file)
    query_adapter = QueryAdapter(config)

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    window = Popup()
    window.show_all()
    window.hide()

    AutoHidePopupDaemon(window).start()
    QueryDaemon(window, query_adapter).start()

    Gtk.main()


def main():
    parser = argparse.ArgumentParser(description='划词翻译')
    parser.add_argument('--debug', default=False, action='store_true', help='调试模式')
    parser.add_argument('--config', help='配置文件，默认 ~/.config/popup-dict/config.ini, 或 /etc/popup-dict/config.ini',
                        metavar='CONFIG_FILE')
    args = parser.parse_args()

    if args.config and not os.path.exists(args.config):
        print("Config file {} does not exist!".format(repr(args.config)), file=sys.stderr)
        sys.exit(1)

    if is_running():
        print('An instance is already running!', file=sys.stderr)
        sys.exit(2)
    else:
        write_pid(os.getpid())
        try:
            start(config_file=args.config, debug=args.debug)
        except ConfigError as e:
            print(e.args[0], file=sys.stderr)
            sys.exit(1)


if __name__ == '__main__':
    main()
